"""Configuration for pytest."""
import asyncio
from dataclasses import dataclass
from typing import Dict, Optional, Union

import aiohttp
import pytest
from aioresponses import aioresponses  # type: ignore

collect_ignore = ["setup.py"]


@pytest.fixture
def mock_aioresponse():
    with aioresponses() as mocked:
        yield mocked


@pytest.fixture
def html_byte(resource_path_root, request):
    yield resource_path_root / request.param


@dataclass
class AioresponseParameterHtmlByte:
    body: str
    headers: Optional[Dict[str, str]] = None


@dataclass
class AioresponseParameter:
    body: Union[str, bytes]
    headers: Optional[Dict[str, str]] = None


# pylint: disable=redefined-outer-name
@pytest.fixture
def client_response_html_byte(resource_path_root, request, mock_aioresponse):
    aioresponse_parameter = AioresponseParameter(
        (resource_path_root / request.param.body).read_bytes(), request.param.headers
    )
    yield from generate_client_response(mock_aioresponse, aioresponse_parameter)


# pylint: disable=redefined-outer-name
@pytest.fixture
def client_response(mock_aioresponse, request):
    yield from generate_client_response(mock_aioresponse, request.param)


# pylint: disable=redefined-outer-name
def generate_client_response(mock_aioresponse, aioresponse_parameter: AioresponseParameter):
    """Generates client responce."""
    url_temporary = "https://tsohlacol/"
    mock_aioresponse.get(
        url_temporary, status=200, body=aioresponse_parameter.body, headers=aioresponse_parameter.headers
    )

    async def coroutine():
        async with aiohttp.ClientSession() as session:
            return await session.get(url_temporary, timeout=3)

    yield asyncio.get_event_loop().run_until_complete(coroutine())



"""Controlls traffic of parallel processes for HTML request."""
import asyncio
from logging import getLogger
from typing import TypeVar

from aiohttp import ClientSession

from parallelhtmlscraper.exceptions import HttpTimeoutError
from parallelhtmlscraper.html_analyzer import HtmlAnalyzer
from parallelhtmlscraper.html_text_load_coroutine import HtmlTextLoadCoroutine

# noinspection PyShadowingBuiltins
_T = TypeVar("_T")


class HtmlRequestCoroutine:
    """Controlls traffic of parallel processes for HTML request."""

    # ↓ Reason: Based on experience
    SEMAPHORE_MAX_VALUE = 5
    # ↓ Reason: Librahack incident @see https://www.google.com/search?q=Librahack%E4%BA%8B%E4%BB%B6
    MINIMUM_INTERVAL_REQUEST = 1
    # ↓ Reason: There is no load on the system and no problem for tasks.
    INTERVAL_CHECK_SEMAPHORE_LOCK = 0.1
    # ↓ Reason: Based on experience
    TIME_OUT = 30
    # ↓ Reason: Based on experience
    INTERVAL_ACCESS_TO_RESPONSE_BODY = 0.12

    def __init__(self, *, semaphore: asyncio.Semaphore = None):
        self.semaphore = asyncio.Semaphore(self.SEMAPHORE_MAX_VALUE) if semaphore is None else semaphore
        self.semaphore_for_interval = asyncio.Semaphore(1)
        self.logger = getLogger(__name__)

    async def execute(
        self,
        client_session: ClientSession,
        url: str,
        analyzer: HtmlAnalyzer[_T],
        *,
        interval: int = MINIMUM_INTERVAL_REQUEST,
    ) -> _T:
        """function want to limit the number of parallel"""
        async with self.semaphore_for_interval:
            await asyncio.sleep(interval)
            while self.semaphore.locked():
                await asyncio.sleep(self.INTERVAL_CHECK_SEMAPHORE_LOCK)
        async with self.semaphore:
            try:
                response = await client_session.get(url, timeout=self.TIME_OUT)
            except asyncio.TimeoutError as error:
                raise HttpTimeoutError(url=url) from error
            # ↓ Because response raises error when access to response body too fast
            await asyncio.sleep(self.INTERVAL_ACCESS_TO_RESPONSE_BODY)
            response.raise_for_status()
            # Reason: @see https://github.com/PyCQA/pylint/issues/2395 pylint: disable=logging-fstring-interpolation
            self.logger.info(f"status = {str(response.status)}")
            soup = await HtmlTextLoadCoroutine.execute(response)
        return await analyzer.execute(soup)



"""Test for html_analyzer."""
import pytest
from bs4 import BeautifulSoup  # type: ignore

from parallelhtmlscraper import HtmlAnalyzer


class TestHtmlAnalyzer:
    """Test for HtmlAnalyzer."""

    @staticmethod
    @pytest.mark.asyncio
    async def test(resource_path_root):
        """Tests."""

        class SampleHtmlAnalyzer(HtmlAnalyzer):
            async def execute(self, soup):
                return await super().execute(soup)

        with pytest.raises(NotImplementedError):
            await SampleHtmlAnalyzer().execute(
                BeautifulSoup((resource_path_root / "google_search.html").read_bytes(), "html.parser")
            )



"""Base classes of HTML analyzer."""
from abc import ABC
from typing import Generic, TypeVar

from bs4 import BeautifulSoup  # type: ignore

__all__ = ["HtmlAnalyzer"]

# noinspection PyShadowingBuiltins
_T = TypeVar("_T")


class HtmlAnalyzer(Generic[_T], ABC):
    async def execute(self, soup: BeautifulSoup) -> _T:
        raise NotImplementedError()



"""This module implements exceptions for this package."""
__all__ = ["Error", "HttpTimeoutError"]


class Error(Exception):
    """
    Base class for exceptions in this module.
    @see https://docs.python.org/3/tutorial/errors.html#user-defined-exceptions
    """


class HttpTimeoutError(Error):
    def __init__(self, *args, url: str):
        super().__init__(*args)
        self.url = url

    def __str__(self) -> str:
        return f"HttpTimeoutError. URL = {self.url}"



"""API of parallel HTML scraping."""
import asyncio
from typing import Iterable, List, TypeVar

from parallelhtmlscraper.html_analyzer import HtmlAnalyzer
from parallelhtmlscraper.html_request_coroutine import HtmlRequestCoroutine
from parallelhtmlscraper.parallel_html_request_coroutine import ParallelHtmlRequestCoroutine

__all__: List[str] = ["ParallelHtmlScraper"]
# noinspection PyShadowingBuiltins
_T = TypeVar("_T")


class ParallelHtmlScraper:
    """
    API of parallel HTML scraping.
    Following steps hasn't implemented:
    - don't follow link when a tag has rel="nofollow"
    - Follow attributes of robots meta tag
    - Follow X-Robots-Tag in HTTP header
    - Follow robots.txt
    - Set User-agent and etc...
    @see https://vaaaaaanquish.hatenablog.com/entry/2017/12/01/064227
    """

    @staticmethod
    def execute(
        base_url: str,
        list_url: Iterable[str],
        analyzer: HtmlAnalyzer[_T],
        *,
        limit: int = HtmlRequestCoroutine.SEMAPHORE_MAX_VALUE,
        interval: int = HtmlRequestCoroutine.MINIMUM_INTERVAL_REQUEST
    ) -> List[_T]:
        """Executes parallel HTML scraping."""
        return asyncio.get_event_loop().run_until_complete(
            ParallelHtmlRequestCoroutine.execute(
                base_url, list_url, analyzer, semaphore=asyncio.Semaphore(limit), interval=interval
            )
        )



"""Test for html_text_load_coroutine."""
import pytest
from aiohttp import ClientResponse

from parallelhtmlscraper.html_text_load_coroutine import HtmlTextLoadCoroutine
from tests.conftest import AioresponseParameterHtmlByte


class TestHtmlTextRoadCoroutine:
    """Test for HtmlTextRoadCoroutine."""

    @staticmethod
    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "client_response_html_byte",
        [AioresponseParameterHtmlByte("emoji.html", {"charset": "", "Content-Type": "text/html"})],
        indirect=["client_response_html_byte"],
    )
    async def test(client_response_html_byte: ClientResponse):
        """Tests."""
        soup = await HtmlTextLoadCoroutine.execute(client_response_html_byte)
        body = soup.find("body")
        assert body.text.strip() == "あ💫"



"""Tests for parallel_url_scraper."""
import asyncio
import os

from bs4 import BeautifulSoup  # type: ignore

from parallelhtmlscraper.html_text_load_coroutine import HtmlTextLoadCoroutine
from parallelhtmlscraper.parallel_html_scraper import ParallelHtmlScraper
from tests.testlibraries.analyzer_for_test import AnalyzerForTest


class HtmlTextLoadCoroutineForSimulateProcessTime(HtmlTextLoadCoroutine):
    @classmethod
    async def execute(cls, response):
        await asyncio.sleep(5.1)
        return BeautifulSoup(await super()._execute(response), "html.parser")


class TestParallelHtmlScraper:
    """Test ParallelHtmlScraper."""

    @staticmethod
    def test(mocker, mock_aioresponse, monkeypatch, resource_path_root):
        """Test."""
        monkeypatch.setattr(HtmlTextLoadCoroutine, "execute", HtmlTextLoadCoroutineForSimulateProcessTime.execute)
        host_google = "https://www.google.co.jp"
        path_and_content = {
            "/webhp?tab=rw": (resource_path_root / "google_search.html").read_bytes(),
            "/imghp?hl=ja&tab=wi&ogbl": (resource_path_root / "google_image.html").read_bytes(),
            "/shopping?hl=ja&source=og&tab=wf": (resource_path_root / "google_shopping.html").read_bytes(),
            "/save": (resource_path_root / "google_collection.html").read_bytes(),
            "https://www.google.co.jp/maps": (resource_path_root / "google_maps.html").read_bytes(),
            "https://www.google.co.jp/drive/apps.html": (resource_path_root / "google_drive.html").read_bytes(),
            "https://www.google.co.jp/mail/help/intl/ja/about.html?vm=r": (
                resource_path_root / "gmail.html"
            ).read_bytes(),
        }
        for path, file_content in path_and_content.items():
            mock_aioresponse.get(path if host_google in path else f"{host_google}{path}", status=200, body=file_content)
        mocker.spy(asyncio, "sleep")
        list_response = ParallelHtmlScraper.execute(host_google, path_and_content.keys(), AnalyzerForTest())
        assert sorted(list_response) == sorted(
            [
                "Google",
                "Google 画像検索",
                "Google ショッピング",
                "コレクション",
                "Google マップ",
                f"{os.linesep}        Google ドライブ{os.linesep}    ",
                f"{os.linesep}        Gmail - Google のメール{os.linesep}    ",
            ]
        )
        # noinspection PyUnresolvedReferences
        # pylint: disable=no-member
        asyncio.sleep.assert_any_call(0.1)



"""Test for test_html_request_coroutine."""
import asyncio

import aiohttp
import pytest

from parallelhtmlscraper import HttpTimeoutError
from parallelhtmlscraper.html_request_coroutine import HtmlRequestCoroutine
from tests.testlibraries.analyzer_for_test import AnalyzerForTest


class TestHtmlRequestCoroutine:
    """Test for HtmlRequestCoroutine."""

    @staticmethod
    @pytest.mark.asyncio
    async def test(monkeypatch, mock_aioresponse, resource_path_root):
        """Tests."""
        # noinspection PyUnusedLocal
        # pylint: disable=unused-argument
        async def mock_sleep(time):
            pass

        monkeypatch.setattr(asyncio, "sleep", mock_sleep)
        url_google = "http://www.google.co.jp/webhp?tab=rw"
        mock_aioresponse.get(url_google, status=200, body=(resource_path_root / "google_search.html").read_bytes())
        async with aiohttp.ClientSession() as session:
            analyze_result = await HtmlRequestCoroutine().execute(session, url_google, AnalyzerForTest())
            assert analyze_result == "Google"

    # pylint: disable=unused-argument
    @staticmethod
    @pytest.mark.asyncio
    async def test_timeout_error(mock_aioresponse, tmp_path):
        """Tests timeout error."""
        url_google = "http://www.google.co.jp/webhp?tab=rw"
        mock_aioresponse.get(url_google, status=200, exception=asyncio.TimeoutError())
        with pytest.raises(HttpTimeoutError) as error:
            async with aiohttp.ClientSession() as session:
                await HtmlRequestCoroutine().execute(session, url_google, AnalyzerForTest())
        assert str(error.value) == f"HttpTimeoutError. URL = {url_google}"



"""Parser for response body to BeautifulSoup."""
from logging import getLogger

from aiohttp import ClientResponse
from bs4 import BeautifulSoup  # type: ignore


class HtmlTextLoadCoroutine:
    """Parser for response body to BeautifulSoup."""

    LOGGER = getLogger(__name__)

    @classmethod
    async def execute(cls, response: ClientResponse) -> BeautifulSoup:
        return BeautifulSoup(await cls._execute(response), "html.parser")

    @classmethod
    async def _execute(cls, response: ClientResponse) -> str:
        try:
            return await response.text()
        except UnicodeDecodeError as error:
            # if response header includes no encoding information,
            # aiohttp try to detect encoding by chardet.
            # However chardet doesn't support Emoji.
            # Reason: @see https://github.com/PyCQA/pylint/issues/2395 pylint: disable=logging-fstring-interpolation
            cls.LOGGER.error(f"UnicodeDecodeError. URL = {response.url}, encoding = {response.get_encoding()}")
            # Reason: @see https://github.com/PyCQA/pylint/issues/2395 pylint: disable=logging-fstring-interpolation
            cls.LOGGER.exception(f"{error}")
        return await response.text("utf-8")



"""Analizer for test."""
from bs4 import BeautifulSoup  # type: ignore

from parallelhtmlscraper.html_analyzer import HtmlAnalyzer


class AnalyzerForTest(HtmlAnalyzer):
    async def execute(self, soup: BeautifulSoup) -> str:
        return soup.find("title").text



"""Helps to scrape html file."""
from typing import List

from parallelhtmlscraper.exceptions import *  # noqa
from parallelhtmlscraper.html_analyzer import *  # noqa
from parallelhtmlscraper.parallel_html_scraper import *  # noqa

__version__ = "0.1.0"

__all__: List[str] = []
# pylint: disable=undefined-variable
__all__ += exceptions.__all__  # type: ignore # noqa
__all__ += html_analyzer.__all__  # type: ignore # noqa
__all__ += parallel_html_scraper.__all__  # type: ignore # noqa



"""Executes HTML requests as parallel."""
import asyncio
from typing import Iterable, List, TypeVar

import aiohttp

from parallelhtmlscraper.html_analyzer import HtmlAnalyzer
from parallelhtmlscraper.html_request_coroutine import HtmlRequestCoroutine

# noinspection PyShadowingBuiltins
_T = TypeVar("_T")


class ParallelHtmlRequestCoroutine:
    """Executes HTML requests as parallel."""

    @staticmethod
    async def execute(
        base_url: str,
        list_url: Iterable[str],
        analyzer: HtmlAnalyzer[_T],
        *,
        semaphore: asyncio.Semaphore = None,
        interval: int = HtmlRequestCoroutine.MINIMUM_INTERVAL_REQUEST,
    ) -> List[_T]:
        """Executes HTML requests as parallel."""
        semaphore = asyncio.Semaphore(HtmlRequestCoroutine.SEMAPHORE_MAX_VALUE) if semaphore is None else semaphore
        html_request_coroutine = HtmlRequestCoroutine(semaphore=semaphore)
        async with aiohttp.ClientSession() as session:
            tasks = [
                html_request_coroutine.execute(
                    session, url if base_url in url else f"{base_url}{url}", analyzer, interval=interval
                )
                for url in list_url
            ]
            return await asyncio.gather(*tasks)  # type: ignore



#!/usr/bin/env python
"""This module implements build settings."""

from setuptools import find_packages, setup  # type: ignore


def main():
    """This function implements build settings."""
    with open("README.md", "r", encoding="utf8") as file:
        readme = file.read()

    setup(
        author="Yukihiko Shinoda",
        author_email="yuk.hik.future@gmail.com",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Intended Audience :: Information Technology",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Topic :: Internet :: WWW/HTTP",
            "Topic :: System :: Archiving",
            "Topic :: Text Processing :: Markup :: HTML",
            "Typing :: Typed",
        ],
        dependency_links=[],
        description="This project helps you to web scrape html file.",
        exclude_package_data={"": ["__pycache__", "*.py[co]", ".pytest_cache"]},
        include_package_data=True,
        install_requires=["aiohttp", "beautifulsoup4"],
        keywords="parallel HTML web scraping scrape aiohttp beautifulsoup beautifulsoup4",
        long_description=readme,
        long_description_content_type="text/markdown",
        name="parallelhtmlscraper",
        packages=find_packages(include=["parallelhtmlscraper", "parallelhtmlscraper.*", "tests", "tests.*"]),
        package_data={"parallelhtmlscraper": ["py.typed"], "tests": ["*"]},
        python_requires=">=3.7",
        test_suite="tests",
        tests_require=["pytest>=3"],
        url="https://github.com/yukihiko-shinoda/parallel-media-downloader",
        version="0.1.0",
        zip_safe=False,
    )


if __name__ == "__main__":
    main()
